"""

Pylp decorators.

Copyright (C) 2017 The Pylp Authors.
This file is under the MIT License.

"""

import sys
from pylp.lib.tasks import task as __task


# Export this module as 'pylp.fn'
sys.modules['pylp.fn'] = sys.modules[__name__]


# Create a task
def task(name, deps = None):
	def __decorated(func):
		__task(name, deps, func)
		return func
	return __decorated



"""

Pylp API.

Copyright (C) 2017 The Pylp Authors.
This file is under the MIT License.

"""

__version__ = "0.1.3"

from pylp.lib.tasks import task
from pylp.lib.tasks import start

from pylp.lib.src import src
from pylp.lib.dest import dest

from pylp.lib.utils import pipes
import pylp.lib.decorators as fn

from pylp.lib.stream import Stream
from pylp.lib.transformer import Transformer
from pylp.lib.file import File



"""

Some useful functions.

Copyright (C) 2017 The Pylp Authors.
This file is under the MIT License.

"""


# Pipe several transformers end to end
def pipes(stream, *transformers):
    for transformer in transformers:
        stream = stream.pipe(transformer)
    return stream


# Get a representative text of a time (in s)
def time_to_text(time):
    if time < 1:
        return str(round(time * 1000)) + " ms"
    else:
        return str(round(time, 1)) + " s"



"""

Transform a stream to another.

Copyright (C) 2017 The Pylp Authors.
This file is under the MIT License.

"""

import asyncio


# The transformr class
class Transformer():

	# Constructor
	def __init__(self):
		self.stream = None


	# Append a new file in the stream
	def append(self, file):
		self.stream.next.append_file(file)


	# Function called when a file need to be transformed
	async def transform(self, file):
		pass

	# Function called when all files are transformed
	async def flush(self):
		pass



"""

Change the color of the terminal.

Copyright (C) 2017 The Pylp Authors.
This file is under the MIT License.

"""

import os
import ctypes


# Get the current system
is_win32 = (os.name == "nt")

# If we are on Windows
if is_win32:
	handle = ctypes.windll.kernel32.GetStdHandle(-11)
	last_fg = 7
	last_bg = 0



# Supported foreground colors, ordered by [ansi, win32]
foreground_colors = {
	"default": [39, 7],
	"black": [30, 0],
	"red": [31, 4],
	"green": [32, 2],
	"yellow": [33, 6],
	"blue": [34, 1],
	"magenta": [35, 5],
	"cyan": [36, 3],
	"light-gray": [37, 7],
	"dark-gray": [90, 8],
	"light-red": [91, 12],
	"light-green": [92, 10],
	"light-yellow": [93, 14],
	"light-blue": [94, 9],
	"light-magenta": [95, 13],
	"light-cyan": [96, 11],
	"white": [97, 15]
}

# Supported background colors, with [ansi, win32]
background_colors = {
	"default": [49, 0],
	"black": [40, 0],
	"red": [41, 64],
	"green": [42, 32],
	"yellow": [43, 96],
	"blue": [44, 16],
	"magenta": [45, 80],
	"cyan": [46, 48],
	"light-gray": [47, 112],
	"dark-gray": [100, 128],
	"light-red": [101, 192],
	"light-green": [102, 160],
	"light-yellow": [103, 224],
	"light-blue": [104, 144],
	"light-magenta": [105, 208],
	"light-cyan": [106, 176],
	"white": [107, 240]
}



# Set the foreground color
def foreground(color):
	if color not in foreground_colors:
		return
	if is_win32:
		last_fg = foreground_colors[color][1]
		set_color_win32(last_fg | last_bg)
	else:
		set_color_ansi(foreground_colors[color][0])


# Set the background color
def background(color):
	if color not in background_colors:
		return
	if is_win32:
		last_bg = background_colors[color][1]
		set_color_win32(last_fg | last_bg)
	else:
		set_color_ansi(background_colors[color][0])



# Set the color of the terminal on Windows
def set_color_win32(color):
	ctypes.windll.kernel32.SetConsoleTextAttribute(handle, color)

# Set the color of the terminal on Linux and MacOS
def set_color_ansi(color):
	print("\e[" + str(color) + "m", end='')



"""

Stream with a piping system.

Copyright (C) 2017 The Pylp Authors.
This file is under the MIT License.

"""

import asyncio


# An asynchronous stream containing a transformer
class Stream():

	# Constructor
	def __init__(self, files = None):
		# Files to transform and number of transformed
		self.files = files if files else []
		self.transformed = 0

		# Transformer and next stream
		self.transformer = None
		self.next = None

		# Is the stream ended?
		self.ended = False

		# Events
		self.onpiped = asyncio.Future()



	# Append a new file in the stream
	def append_file(self, file):
		self.files.append(file)

		if self.transformer:
			future = asyncio.ensure_future(self.transformer.transform(file))
			future.add_done_callback(self.handle_transform)


	# Handle a 'transform' callback
	def handle_transform(self, task):
		self.transformed += 1
		file = task.result()

		if file:
			self.next.append_file(file)

		if self.ended and len(self.files) == self.transformed:
			future = asyncio.ensure_future(self.transformer.flush())
			future.add_done_callback(lambda x: self.next.end_of_stream())


	# Tell that no more files will be transformed
	def end_of_stream(self):
		self.ended = True

		if self.next and len(self.files) == self.transformed:
			future = asyncio.ensure_future(self.transformer.flush())
			future.add_done_callback(lambda x: self.next.end_of_stream())



	# Pipe this stream to another
	def pipe(self, transformer):
		if self.next:
			return

		stream = Stream()
		self.next = stream

		self.transformer = transformer
		transformer.stream = self

		for file in self.files:
			future = asyncio.ensure_future(self.transformer.transform(file))
			future.add_done_callback(self.handle_transform)	

		self.onpiped.set_result(None)
		return stream



"""

Log some text into the terminal.

Copyright (C) 2017 The Pylp Authors.
This file is under the MIT License.

"""

import time
import pylp.cli.colors as colors


# Print a message with a color
def cprint(text):
	print(text, end='', flush=colors.is_win32)


# Do we have reset the line?
have_reset = True

# Reset the color and add a new line
def reset(newLine = True):
	global have_reset
	colors.foreground("default")
	have_reset = True
	if newLine:
		print()


# Log the current time (used as prefix in all logs)
def log_time():
	global have_reset
	if have_reset:
		colors.foreground("default")
		cprint("[")
		colors.foreground("dark-gray")
		cprint(time.strftime("%H:%M:%S"))
		colors.foreground("default")
		cprint("] ")
		have_reset = False


# Log a normal text
def log(text):
	log_time()
	colors.foreground("default")
	cprint(text)

# Log an information
def info(text):
	log_time()
	colors.foreground("magenta")
	cprint(text)

# Log an name
def name(text, newLine = False):
	log_time()
	colors.foreground("cyan")
	cprint(text)

# Log an error
def error(text, newLine = False):
	log_time()
	colors.foreground("red")
	cprint(text)



"""

Store the content of a file.

Copyright (C) 2017 The Pylp Authors.
This file is under the MIT License.

"""

import os
import asyncio
from concurrent.futures import ThreadPoolExecutor
from pylp.lib.transformer import Transformer


# The file class
class File():

	# Constructor
	def __init__(self, path, **options):
		# Path of the file
		self.path = os.path.abspath(path)
		self.relpath = None
		self.cwd = options.get('cwd')
		self.relative = os.path.relpath(path, self.cwd)

		# Contents of the file
		self.contents = options.get("contents", "")

		# Options
		self.base = options.get("base")




# Read a local file
class FileReader(Transformer):

	# Constructor
	def __init__(self):
		super().__init__()

		self.exe = ThreadPoolExecutor()
		self.loop = asyncio.get_event_loop()


	# Read a file
	def read_file(self, path):
		with open(path, "r") as file:
			return file.read()


	# Function called when a file need to be transformed
	async def transform(self, file):
		file.contents = await self.loop.run_in_executor(self.exe, self.read_file, file.path)
		return file



"""

Run a pylpfile.

Copyright (C) 2017 The Pylp Authors.
This file is under the MIT License.

"""

import runpy, os, sys
import traceback
import asyncio
import pylp, pylp.cli.logger as logger


# Run a pylpfile
def run(path, tasks):
	# Test if the pylpfile exists
	is_file = os.path.isfile(path)

	if not os.path.isfile(path):
		logger.error("Can't read pylpfile ")
	else:
		logger.log("Using pylpfile ")

	logger.info(path)
	logger.reset()

	if not is_file:
		return


	# Run the pylpfile
	try:
		runpy.run_path(path, None, "pylpfile")
	except Exception as e:
		traceback.print_exc(file=sys.stdout)
		logger.reset()
		logger.error("An error has occurred during the execution of the pylpfile")
		logger.reset()
		return


	# Start the tasks
	for name in tasks:
		pylp.start(name)

	# Wait until all task are executed
	loop = asyncio.get_event_loop()
	loop.run_until_complete(wait_and_quit(loop))



# Wait until all task are executed 
async def wait_and_quit(loop):
	from pylp.lib.tasks import running
	await asyncio.wait(map(lambda runner: runner.future, running))



"""

Some classes for running a task.

Copyright (C) 2017 The Pylp Authors.
License GPL3

"""

import asyncio
import time
import pylp
import pylp.cli.logger as logger
from pylp.lib.transformer import Transformer
from pylp.lib.utils import time_to_text



# Wait until a 'end-of-stream'
class TaskEndTransformer(Transformer):

	# Constructor
	def __init__(self, future):
		super().__init__()
		self.future = future

	# Function called after when files are transformed.
	async def flush(self):
		self.future.set_result(None)




# Execute a task
class TaskRunner():

	# Constructor
	def __init__(self, name, fn, deps, called):
		self.name = name
		self.called = called if called else []
		self.fn = fn

		self.log_starting()	

		if deps:
			self.future = self.start_deps(deps)
		else:
			self.future = self.call_task_fn()


	# Log that the task has started
	def log_starting(self):
		self.start_time = time.perf_counter()

		logger.log("Starting '")
		logger.name(self.name)
		logger.log("'...")
		logger.reset()


	# Log that this task is done
	def log_finished(self):
		delta = time.perf_counter() - self.start_time

		logger.log("Finished '")
		logger.name(self.name)
		logger.log("' after ")
		logger.info(time_to_text(delta))
		logger.reset()


	# Call the function attached to the task
	def call_task_fn(self):
		if not self.fn:
			self.log_finished()
		else:
			future = asyncio.Future()
			future.add_done_callback(lambda x: self.log_finished())
			self.fn().pipe(TaskEndTransformer(future))
			return future


	# Start running dependencies
	async def start_deps(self, deps):
		# Get only new dependencies
		deps = list(filter(lambda dep: dep not in self.called, deps))
		self.called += deps

		# Start only existing dependencies
		runners = list(filter(lambda x: x, map(lambda dep: pylp.start(dep), deps)))
		if len(runners) != 0:
			await asyncio.wait(map(lambda runner: runner.future, runners))

		# Call the attached function 
		future = self.call_task_fn()
		if future:
			await future



"""

Create and run tasks.

Copyright (C) 2017 The Pylp Authors.
License GPL3

"""

import pylp.cli.logger as logger
from pylp.lib.runner import TaskRunner



# List of created tasks
tasks = {}

# Define a new task.
def task(name, deps = None, fn = None):
	if callable(deps):
		fn = deps
		deps = None

	if not deps and not fn:
		logger.error("The task '%s' is empty" % name)
		logger.reset()
	else:
		tasks[name] = [fn, deps]



# List of running task
running = []

# Start a task
def start(name, called = None):
	if name not in tasks:
		logger.error("Task '%s' not in your pylpfile" % name)
		logger.reset()
	else:
		task = tasks[name]
		runner = TaskRunner(name, task[0], task[1], called)
		running.append(runner)
		return runner



"""

Create a starting stream from local files.

Copyright (C) 2017 The Pylp Authors.
This file is under the MIT License.

"""

import glob
import os
import pylp.lib.config as config
from pylp.lib.file import File, FileReader
from pylp.lib.stream import Stream


# Parse a glob
def parse_glob(path, include, exclude):
	files = glob.glob(path, recursive=True)

	is_neg = path.startswith("!")
	array = exclude if is_neg else include

	for file in files:
		if file not in array:
			array.append(file)



# Remove excluded file from 'include' array
def apply_exclude(include, exclude):
	for file in exclude:
		if file in include:
			include.remove(file)



# Read some files and return a stream
def src(globs, **options):
	# Create an array of globs if only one string is given
	if isinstance(globs, str):
		globs = [ globs ]

	# List of files to include and exclude
	include = []
	exclude = []

	# Find files
	for path in globs:
		parse_glob(path, include, exclude)
	apply_exclude(include, exclude)

	# Create a stream
	stream = Stream()
	options["cwd"] = config.cwd

	# Create a File object for each file to include
	for file in include:
		file = File(file, **options)
		file.relpath = file.path
		stream.append_file(file)

	# No more files to add
	stream.end_of_stream()

	# Pipe a file reader and return the stream
	if options.get("read", True):
		return stream.pipe(FileReader())
	return stream



"""

Write contents to local files.

Copyright (C) 2017 The Pylp Authors.
License GPL3

"""

import os
import asyncio
from concurrent.futures import ThreadPoolExecutor
from pylp.lib.transformer import Transformer



# Return a transformer for writing files
def dest(path, **options):
	return FileWriter(path, **options)



# Get the writing path of a file
def get_path(dest, file, cwd = None):
	if callable(dest):
		return dest(file)

	if not cwd:
		cwd = file.cwd
	if not os.path.isabs(dest):
		dest = os.path.join(cwd, dest)

	if file.base:
		relative = os.path.relpath(file.path, file.base)
	else:
		relative = os.path.basename(file.path)

	return os.path.join(dest, relative)


# Write a file
def write_file(path, contents):
	os.makedirs(os.path.dirname(path), exist_ok=True)
	with open(path, "w") as file:
		file.write(contents)



# Transformer that save local files.
class FileWriter(Transformer):

	# Constructor
	def __init__(self, path, **options):
		super().__init__()

		self.dest = path
		self.cwd = options.get('cwd')

		self.exe = ThreadPoolExecutor()
		self.loop = asyncio.get_event_loop()


	# Function called when a file need to be transformed
	async def transform(self, file):
		path = get_path(self.dest, file)
		await self.loop.run_in_executor(self.exe, write_file, path, file.contents)



"""

Setup Pylp.

Copyright (C) 2017 The Pylp Authors.
This file is under the MIT License.

"""

from setuptools import setup, find_packages
from pylp import __version__ as version


setup(
	name = "pylp",
	version = version,
	author = "Guillaume Gonnet",
	author_email = "gonnet.guillaume97@gmail.com",
	description = "A Python task runner inspired by gulp.js",
	license = "MIT",
	keywords = "pylp build task runner gulp",
	url = "https://github.com/pylp/pylp",
	packages = find_packages(),
	long_description = open("README.rst").read(),
	entry_points = {
		"console_scripts" : ["pylp = pylp.cli.cli:launch_cli",]
	},
	classifiers = [
		"Development Status :: 3 - Alpha",
		"Topic :: Utilities",		
		"Topic :: Software Development :: Build Tools",
		"Framework :: AsyncIO",
		"Programming Language :: Python :: 3 :: Only",
		"Programming Language :: Python :: 3.5",
		"Intended Audience :: Developers",
		"Natural Language :: English",
		"License :: OSI Approved :: MIT License",
	],
)



"""

Start Pylp via the CLI.

Copyright (C) 2017 The Pylp Authors.
This file is under the MIT License.

"""

# Add 'pylp' into 'sys.path' if Pylp is not in python Lib. 
import os, sys
import os.path as path

relpath = path.dirname(path.realpath(__file__))
parent = path.normpath(path.realpath(path.join(relpath, "..")))

if not parent.endswith(path.normpath("Lib/site-package")):
	base = os.path.join(os.path.dirname(__file__), "../..")
	sys.path.insert(0, os.path.abspath(base))


# We can now import all Pylp files
import argparse
from pylp import __version__ as version
import pylp.lib.config as config
from pylp.cli.run import run



# Launch the CLI
def launch_cli():
	# Create the CLI argument parser
	parser = argparse.ArgumentParser(
		prog="pylp",
		description="Call some tasks defined in your pylpfile."
	)


	# Version of Pylp
	parser.add_argument("-v", "--version",
		action="version",
		version="Pylp %s" % version,
		help="get the Pylp version and exit"
	)

	# Set the pylpfile location
	parser.add_argument('--pylpfile',
		nargs=1,
		help="manually set path of pylpfile",
		metavar="<path>"
	)

	# Set the pylpfile location
	parser.add_argument('--cwd',
		nargs=1,
		help="manually set the CWD",
		metavar="<dir path>"
	)

	# List of tasks to execute
	parser.add_argument('tasks',
		nargs="*",
		default=["default"],
		help="tasks to execute (if none, execute the 'default' task)",
		metavar="<task>"
	)


	# Parse the CLI arguments
	args = parser.parse_args()


	# Current working directory (CWD)
	if args.cwd:
		config.cwd = args.cwd
	else:
		config.cwd = os.getcwd()

	# Get the pylpfile location
	pylpfile = args.pylpfile
	if not pylpfile:
		pylpfile = path.join(config.cwd, "pylpfile.py")
	elif not args.cwd:
		config.cwd = path.dirname(pylpfile)


	# Execute the pylpfile
	run(pylpfile, args.tasks)



# Launch the CLI if this is executed directly
if __name__ == "__main__":
    launch_cli()
